from collections import OrderedDict
from logging import Logger
from typing import Any, Dict, Generator, List

import requests
from requests.utils import requote_uri

from the_census._api.interface import (
    ICensusApiFetchService,
    ICensusApiSerializationService,
)
from the_census._api.models import GeographyItem
from the_census._config import Config
from the_census._exceptions import CensusDoesNotExistException, InvalidQueryException
from the_census._geographies.models import GeoDomain
from the_census._utils.chunk import chunk
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._variables.models import Group, GroupVariable, VariableCode

# we can query only 50 variables at a time, max
MAX_QUERY_SIZE = 50
API_URL_FORMAT = "https://api.census.gov/data/{0}/{1}/{2}"


class CensusApiFetchService(ICensusApiFetchService):
    _url: str
    _parser: ICensusApiSerializationService
    _config: Config
    _logger: Logger

    def __init__(
        self,
        config: Config,
        parser: ICensusApiSerializationService,
        loggingFactory: ILoggerFactory,
    ) -> None:
        self._url = API_URL_FORMAT.format(
            config.year, config.datasetType, config.surveyType
        )
        self._parser = parser
        self._config = config
        self._logger = loggingFactory.getLogger(__name__)

    def healthCheck(self) -> None:
        res = requests.get(self._url + ".json")  # type: ignore

        if res.status_code in [404, 400]:
            msg = f"Data does not exist for dataset={self._config.datasetType}; survey={self._config.surveyType}; year={self._config.year}"

            self._logger.exception(msg)

            raise CensusDoesNotExistException(msg)

        self._logger.debug("healthCheck OK")

    @timer
    def geographyCodes(
        self, forDomain: GeoDomain, inDomains: List[GeoDomain] = []
    ) -> Any:

        forClause = f"for={forDomain}"
        inClauses = "&in=".join([str(parent) for parent in inDomains])

        querystring = f"?get=NAME&{forClause}"
        if len(inDomains):
            querystring += f"&in={inClauses}"  # type: ignore

        uriQuerystring: str = requote_uri(querystring)

        return self._fetch(route=uriQuerystring)

    @timer
    def groupData(self) -> Dict[str, Group]:
        groupsRes: Dict[str, List[Dict[str, str]]] = self._fetch(route="/groups.json")

        return self._parser.parseGroups(groupsRes)

    @timer
    def supportedGeographies(self) -> OrderedDict[str, GeographyItem]:
        geogRes = self._fetch(route="/geography.json")

        return self._parser.parseSupportedGeographies(geogRes)

    @timer
    def variablesForGroup(self, group: str) -> List[GroupVariable]:
        res = self._fetch(route=f"/groups/{group}.json")

        return self._parser.parseGroupVariables(res)

    @timer
    def allVariables(self) -> List[GroupVariable]:
        res = self._fetch("/variables.json")

        return self._parser.parseGroupVariables(res)

    @timer
    def stats(
        self,
        variablesCodes: List[VariableCode],
        forDomain: GeoDomain,
        inDomains: List[GeoDomain] = [],
    ) -> Generator[List[List[str]], None, None]:

        # we need the minus 1 since we're also querying name
        for codes in chunk(variablesCodes, MAX_QUERY_SIZE - 1):
            codeStr = ",".join(codes)
            varStr = "get=NAME" + f",{codeStr}" if len(codeStr) > 0 else ""

            domainStr = "for=" + str(forDomain)
            inDomainStr = "&".join([f"in={domain}" for domain in inDomains])

            if len(inDomainStr) > 0:
                domainStr += "&"  # type: ignore
                domainStr += inDomainStr  # type: ignore

            route = f"?{varStr}&{domainStr}"

            uriRoute: str = requote_uri(route)

            # not doing any serializing here, because this is a bit more
            # complicated (we need to convert the stats to the appropriate
            # data types, [e.g., int, float] further up when we're working
            # with dataFrames; there's no real good way to do it down here)

            yield self._fetch(uriRoute)

    def _fetch(self, route: str = "") -> Any:
        ampersandOrQuestionMark = "&" if "?" in route else "?"
        url = self._url + route + ampersandOrQuestionMark + "key=" + self._config.apiKey
        res = requests.get(url)  # type: ignore
        if res.status_code in [400, 404]:
            msg = f"Could not make query for route `{route}`"
            self._logger.exception(msg)
            raise InvalidQueryException(msg)
        if res.status_code == 204:  # no content
            msg = f"Received no content for query for route {route}"
            self._logger.info(msg)

            return []

        return res.json()  # type: ignore



from abc import abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Generic, List, Optional, TypeVar

from the_census._api.models import GeographyItem
from the_census._geographies.models import GeoDomain
from the_census._variables.models import Group, GroupVariable, VariableCode

T = TypeVar("T")


class ICensusDataTransformer(Generic[T]):
    """
    This takes care of converting parsed API data
    into more meaningfully consumable data (e.g., DataFrames)
    """

    @abstractmethod
    def supportedGeographies(self, supportedGeos: OrderedDict[str, GeographyItem]) -> T:
        ...

    @abstractmethod
    def geographyCodes(self, geoCodes: List[List[str]]) -> T:
        ...

    @abstractmethod
    def groups(self, groups: Dict[str, Group]) -> T:
        ...

    @abstractmethod
    def variables(self, variables: List[GroupVariable]) -> T:
        ...

    @abstractmethod
    def stats(
        self,
        results: List[List[List[str]]],
        typeConversions: Dict[str, Any],
        geoDomains: List[GeoDomain],
        columnHeaders: Optional[Dict[VariableCode, str]],
    ) -> T:
        ...



from dataclasses import dataclass
from functools import cache
from typing import Any, Dict, List, cast

import pandas
import pandas as pd
import requests
from tqdm.notebook import tqdm

URL = "https://api.census.gov/data.json"

# documentation: https://www2.census.gov/programs-surveys/acs/tech_docs/subject_definitions/


@dataclass(frozen=True)
class _DatasetsRes:
    year: int
    dataset: List[str]
    isAggregate: bool
    title: str
    description: str

    @classmethod
    def fromJson(cls, jsonRes: Dict[str, Any]):
        return cls(
            cast(int, jsonRes.get("c_vintage")),
            cast(List[str], jsonRes.get("c_dataset")),
            cast(bool, jsonRes.get("c_isAggregate")),
            cast(str, jsonRes.get("title")),
            cast(str, jsonRes.get("description")),
        )


def listAvailableDataSets() -> pd.DataFrame:
    return __listAvailableDataSets()


@cache
def __listAvailableDataSets() -> pd.DataFrame:
    res: Dict[str, Any] = requests.get(URL).json()  # type: ignore
    datasetDicts: List[Dict[str, str]] = []

    availableDatasets: List[_DatasetsRes] = [
        _DatasetsRes.fromJson(datasetJson) for datasetJson in res["dataset"]
    ]

    for dataset in cast(List[_DatasetsRes], tqdm(availableDatasets)):
        # these won't play nice with the tool
        if not dataset.isAggregate:
            continue

        datasetType = ""
        surveyType = ""
        if len(dataset.dataset) > 0:
            datasetType = dataset.dataset[0]
            if len(dataset.dataset) > 1:
                surveyType = "/".join(dataset.dataset[1:])

        datasetDicts.append(
            cast(
                Dict[str, str],
                dict(
                    year=dataset.year,
                    name=dataset.title,
                    description=dataset.description,
                    datasetType=datasetType,
                    surveyType=surveyType,
                ),
            )
        )

    pandas.set_option("display.max_colwidth", None)  # type: ignore

    return (
        pd.DataFrame(datasetDicts)
        .sort_values(by=["year", "name"], ascending=[False, True])
        .reset_index(drop=True)
    )



from dataclasses import dataclass

CACHE_DIR = "cache"


@dataclass(frozen=True)
class Config:
    """
    Stores basic information so that we don't need to
    pass the same variables around all of the time
    """

    year: int = 2020
    datasetType: str = "acs"
    surveyType: str = "acs1"
    cacheDir: str = CACHE_DIR
    shouldLoadFromExistingCache: bool = False
    shouldCacheOnDisk: bool = False
    replaceColumnHeaders: bool = False
    apiKey: str = ""



from logging import Logger

import pandas as pd

from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._variables.models import GroupCode
from the_census._variables.repository.interface import IVariableRepository
from the_census._variables.search.interface import IVariableSearchService


class VariableSearchService(IVariableSearchService[pd.DataFrame]):
    _variableRepository: IVariableRepository[pd.DataFrame]
    _logger: Logger

    def __init__(
        self,
        variableRepository: IVariableRepository[pd.DataFrame],
        loggerFactory: ILoggerFactory,
    ) -> None:
        self._variableRepository = variableRepository
        self._logger = loggerFactory.getLogger(__name__)

    @timer
    def searchGroups(self, regex: str) -> pd.DataFrame:
        self._logger.debug(f"searching groups for regex: `{regex}`")

        groups = self._variableRepository.getGroups()

        if groups.empty:
            self._logger.info("There are no groups for this dataset")

            return pd.DataFrame()

        series: pd.Series[bool] = groups["description"].str.contains(  # type: ignore
            regex, case=False
        )

        return groups[series].reset_index(
            drop=True,
        )

    @timer
    def searchVariables(
        self,
        regex: str,
        *inGroups: GroupCode,
    ) -> pd.DataFrame:

        self._logger.debug(f"searching variables for pattern `{regex}`")

        variables: pd.DataFrame
        if not len(inGroups):
            variables = self._variableRepository.getAllVariables()
        else:
            variables = self._variableRepository.getVariablesByGroup(*inGroups)

        if variables.empty:
            self._logger.info("There are no variables for this dataset")
            return pd.DataFrame()

        series = variables["name"].str.contains(regex, case=False)  # type: ignore

        return variables[series].reset_index(drop=True)  # type: ignore



# from dataclasses import dataclass
from dataclasses import dataclass, field
from typing import Any, Dict, Literal, NewType

from the_census._utils.cleanVariableName import cleanVariableName

VariableCode = NewType("VariableCode", str)
GroupCode = NewType("GroupCode", str)


@dataclass(frozen=True)
class Group:
    """
    DTO for the group retrieved from the API
    """

    code: GroupCode
    description: str
    variables: str = field(default="")
    cleanedName: str = field(default="")

    @classmethod
    def fromJson(cls, jsonDict: Dict[str, str]):
        code = jsonDict.get("name", "")
        description = jsonDict.get("description", "")
        variables = jsonDict.get("variables", "")

        return cls(
            GroupCode(code), description, variables, cleanVariableName(description)
        )

    @classmethod
    def fromDfRecord(cls, record: Dict[str, Any]):
        return cls(
            GroupCode(record["code"]),
            record["description"],
            cleanedName=record["cleanedName"],
        )


@dataclass
class GroupVariable:
    """
    DTO for a variable retrieved from the API
    """

    code: VariableCode
    groupCode: GroupCode
    groupConcept: str
    name: str
    limit: int = field(default=0)
    predicateOnly: bool = field(default=False)
    predicateType: Literal["string", "int", "float"] = field(default="string")
    cleanedName: str = field(default="")

    @classmethod
    def fromJson(cls, code: str, jsonData: Dict[Any, Any]):
        groupCode = jsonData.get("group", "")
        groupConcept = jsonData.get("concept", "")
        label = jsonData.get("label", "")
        limit = jsonData.get("limit", 0)
        predicateOnly = jsonData.get("predicateOnly", False)
        predicateType = jsonData.get("predicateType", "string")
        cleanedName = cleanVariableName(label)

        return cls(
            VariableCode(code),
            GroupCode(groupCode),
            groupConcept,
            label,
            limit,
            predicateOnly,
            predicateType,
            cleanedName,
        )

    @classmethod
    def fromDfRecord(cls, record: Dict[str, Any]):
        return cls(
            VariableCode(record["code"]),
            GroupCode(record["groupCode"]),
            record["groupConcept"],
            record["name"],
            record["limit"],
            record["predicateOnly"],
            record["predicateType"],
            record["cleanedName"],
        )

    def __hash__(self) -> int:
        return hash(self.code)



from typing import List

import pandas as pd

from the_census._api.fetch import ICensusApiFetchService
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain, SupportedGeoSet
from the_census._stats.interface import ICensusStatisticsService
from the_census._variables.models import GroupCode, VariableCode
from the_census._variables.repository.interface import IVariableRepository
from the_census._variables.repository.models import GroupSet, VariableSet
from the_census._variables.search.interface import IVariableSearchService


class CensusClient:
    """
    The Census client. Using a layer of indirection to hide
    additional properties from consumers.
    """

    _variableRepo: IVariableRepository[pd.DataFrame]
    _variableSearch: IVariableSearchService[pd.DataFrame]
    _stats: ICensusStatisticsService[pd.DataFrame]
    _geoRepo: IGeographyRepository[pd.DataFrame]

    def __init__(
        self,
        variableRepo: IVariableRepository[pd.DataFrame],
        variableSearch: IVariableSearchService[pd.DataFrame],
        stats: ICensusStatisticsService[pd.DataFrame],
        api: ICensusApiFetchService,
        geoRepo: IGeographyRepository[pd.DataFrame],
    ) -> None:
        self._variableRepo = variableRepo
        self._variableSearch = variableSearch
        self._stats = stats
        self._geoRepo = geoRepo

        # if this healthcheck fails, it will throw, and we
        # won't instantiate the client
        api.healthCheck()

    # search
    def searchGroups(self, regex: str) -> pd.DataFrame:
        return self._variableSearch.searchGroups(regex).copy(deep=True)

    def searchVariables(
        self,
        regex: str,
        *inGroups: GroupCode,
    ) -> pd.DataFrame:
        return self._variableSearch.searchVariables(regex, *inGroups).copy(deep=True)

    # repo
    def getGeographyCodes(
        self, forDomain: GeoDomain, *inDomains: GeoDomain
    ) -> pd.DataFrame:
        return self._geoRepo.getGeographyCodes(forDomain, *inDomains).copy(deep=True)

    def getGroups(self) -> pd.DataFrame:
        return self._variableRepo.getGroups().copy(deep=True)

    def getVariablesByGroup(self, *groups: GroupCode) -> pd.DataFrame:
        return self._variableRepo.getVariablesByGroup(*groups).copy(deep=True)

    def getAllVariables(self) -> pd.DataFrame:
        return self._variableRepo.getAllVariables().copy(deep=True)

    def getSupportedGeographies(self) -> pd.DataFrame:
        return self._geoRepo.getSupportedGeographies().copy(deep=True)

    def getStats(
        self,
        variablesToQuery: List[VariableCode],
        forDomain: GeoDomain,
        *inDomains: GeoDomain,
    ) -> pd.DataFrame:
        return self._stats.getStats(variablesToQuery, forDomain, *inDomains).copy(
            deep=True
        )

    # property variables for Jupyter notebook usage

    @property
    def variables(self) -> VariableSet:
        return self._variableRepo.variables

    @property
    def groups(self) -> GroupSet:
        return self._variableRepo.groups

    @property
    def supportedGeographies(self) -> SupportedGeoSet:
        return self._geoRepo.supportedGeographies



from functools import cache
from logging import Logger
from typing import Tuple

import pandas as pd

from the_census._api.interface import ICensusApiFetchService
from the_census._dataTransformation.interface import ICensusDataTransformer
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain, SupportedGeoSet
from the_census._persistence.interface import ICache
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._utils.unique import getUnique

SUPPORTED_GEOS_FILE = "supportedGeographies.csv"


class GeographyRepository(IGeographyRepository[pd.DataFrame]):
    _cache: ICache[pd.DataFrame]
    _api: ICensusApiFetchService
    _transformer: ICensusDataTransformer[pd.DataFrame]
    _logger: Logger

    def __init__(
        self,
        cache: ICache[pd.DataFrame],
        api: ICensusApiFetchService,
        transformer: ICensusDataTransformer[pd.DataFrame],
        loggerFactory: ILoggerFactory,
    ) -> None:
        self._cache = cache
        self._api = api
        self._transformer = transformer
        self._logger = loggerFactory.getLogger(__name__)

        self._supportedGeographies = SupportedGeoSet()

    @timer
    def getGeographyCodes(
        self, forDomain: GeoDomain, *inDomains: GeoDomain
    ) -> pd.DataFrame:
        return self.__getGeographyCodes(
            forDomain, inDomains=tuple(getUnique(inDomains))
        )

    @cache
    def __getGeographyCodes(
        self, forDomain: GeoDomain, inDomains: Tuple[GeoDomain, ...] = ()
    ) -> pd.DataFrame:
        self._logger.debug(f"getting geography codes for {forDomain} in {inDomains}")
        res = self._api.geographyCodes(forDomain, list(inDomains))

        if len(res) == 0:
            return pd.DataFrame()

        df = self._transformer.geographyCodes(res)

        return df

    @timer
    def getSupportedGeographies(self) -> pd.DataFrame:
        return self.__getSupportedGeographies()

    @cache
    def __getSupportedGeographies(self) -> pd.DataFrame:
        self._logger.debug("getting supported geographies")

        df = self._cache.get(SUPPORTED_GEOS_FILE)

        if df is None:
            df = pd.DataFrame()

        if df.empty:
            res = self._api.supportedGeographies()

            if len(res) == 0:
                return pd.DataFrame()

            df = self._transformer.supportedGeographies(res)

            self._cache.put(SUPPORTED_GEOS_FILE, df)

        self._supportedGeographies.add(*df["name"].tolist())

        return df



from dataclasses import dataclass, field

from the_census._utils.cleanVariableName import cleanVariableName


@dataclass(frozen=True)
class GeoDomain:
    name: str
    codeOrWildcard: str = field(default="*")

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"{self.name}:{self.codeOrWildcard}"


class SupportedGeoSet:
    def __init__(self) -> None:
        super().__init__()

    def add(self, *geoName: str):
        mapping = {cleanVariableName(name): name for name in geoName}
        self.__dict__.update(mapping)



import logging


class ModuleFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        modulesToLog = ["", "urllib3"]

        for module in modulesToLog:
            if record.name.startswith(module):
                return True

        return False



from functools import cache
from logging import Logger
from typing import Any, Dict, List, Set, Tuple

import pandas as pd

from the_census._api.interface import ICensusApiFetchService
from the_census._config import Config
from the_census._dataTransformation.interface import ICensusDataTransformer
from the_census._exceptions import EmptyRepositoryException
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain
from the_census._stats.interface import ICensusStatisticsService
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._utils.unique import getUnique
from the_census._variables.models import VariableCode
from the_census._variables.repository.interface import IVariableRepository


class CensusStatisticsService(ICensusStatisticsService[pd.DataFrame]):
    _api: ICensusApiFetchService
    _transformer: ICensusDataTransformer[pd.DataFrame]
    _variableRepo: IVariableRepository[pd.DataFrame]
    _geoRepo: IGeographyRepository[pd.DataFrame]
    _config: Config
    _logger: Logger

    def __init__(
        self,
        api: ICensusApiFetchService,
        transformer: ICensusDataTransformer[pd.DataFrame],
        variableRepo: IVariableRepository[pd.DataFrame],
        geoRepo: IGeographyRepository[pd.DataFrame],
        config: Config,
        loggerFactory: ILoggerFactory,
    ) -> None:
        self._api = api
        self._transformer = transformer
        self._variableRepo = variableRepo
        self._config = config
        self._geoRepo = geoRepo
        self._logger = loggerFactory.getLogger(__name__)

    @timer
    def getStats(
        self,
        variablesToQuery: List[VariableCode],
        forDomain: GeoDomain,
        *inDomains: GeoDomain,
    ) -> pd.DataFrame:

        return self.__getStats(
            variablesToQuery=tuple(getUnique(variablesToQuery)),
            forDomain=forDomain,
            inDomains=tuple(getUnique(inDomains)),
        )

    @cache
    def __getStats(
        self,
        variablesToQuery: Tuple[VariableCode],
        forDomain: GeoDomain,
        inDomains: Tuple[GeoDomain],
    ) -> pd.DataFrame:

        pullStats = lambda: self._api.stats(
            list(variablesToQuery), forDomain, list(inDomains)
        )

        apiResults: List[List[List[str]]] = [res for res in pullStats()]

        columnHeaders, typeConversions = self._getVariableNamesAndTypeConversions(
            set(variablesToQuery)
        )

        sortedGeoDomains = self._sortGeoDomains([forDomain] + list(inDomains))

        df = self._transformer.stats(
            apiResults,
            typeConversions,
            sortedGeoDomains,
            columnHeaders=columnHeaders if self._config.replaceColumnHeaders else None,
        )

        return df

    def _getVariableNamesAndTypeConversions(
        self, variablesToQuery: Set[VariableCode]
    ) -> Tuple[Dict[VariableCode, str], Dict[str, Any]]:

        relevantVariables = {
            variable.code: variable
            for variable in self._variableRepo.variables.values()
            if variable.code in variablesToQuery
        }
        if len(relevantVariables) != len(variablesToQuery):
            msg = f"Queried {len(variablesToQuery)} variables, but found only {len(relevantVariables)} in repository"

            self._logger.exception(msg)

            raise EmptyRepositoryException(msg)

        hasDuplicateNames = len(
            {v.cleanedName for v in relevantVariables.values()}
        ) < len(variablesToQuery)

        typeConversions: Dict[str, Any] = {}
        columnHeaders: Dict[VariableCode, str] = {}
        for k, v in relevantVariables.items():
            if v.predicateType in ["int", "float"]:
                typeConversions.update({k: float})

            cleanedVarName = v.cleanedName
            if hasDuplicateNames:
                cleanedVarName += f"_{v.groupCode}"

            columnHeaders.update({k: cleanedVarName})

        return columnHeaders, typeConversions

    def _sortGeoDomains(self, geoDomains: List[GeoDomain]) -> List[GeoDomain]:
        supportedGeos = self._geoRepo.getSupportedGeographies()
        geoHierarchyMapping: List[Dict[str, str]] = (
            supportedGeos[["name", "hierarchy"]].drop_duplicates().to_dict("records")
        )
        geoNameToHierarchy = {
            mapping["name"]: mapping["hierarchy"] for mapping in geoHierarchyMapping
        }

        return sorted(
            geoDomains, key=lambda geoDomain: geoNameToHierarchy[geoDomain.name]
        )



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from the_census._geographies.models import GeoDomain, SupportedGeoSet

T = TypeVar("T")


class IGeographyRepository(ABC, Generic[T]):
    """
    Gets and stores all geography information
    """

    _supportedGeographies: SupportedGeoSet

    @abstractmethod
    def getSupportedGeographies(self) -> T:
        ...

    @abstractmethod
    def getGeographyCodes(self, forDomain: GeoDomain, *inDomains: GeoDomain) -> T:
        ...

    @property
    def supportedGeographies(self) -> SupportedGeoSet:
        return self._supportedGeographies



# pyright: reportUnknownMemberType=false

import os
from typing import List, cast

import dotenv
import pandas
import punq

from the_census._api.fetch import CensusApiFetchService
from the_census._api.interface import (
    ICensusApiFetchService,
    ICensusApiSerializationService,
)
from the_census._api.serialization import ApiSerializationService
from the_census._client import CensusClient
from the_census._config import CACHE_DIR, Config
from the_census._dataTransformation.interface import ICensusDataTransformer
from the_census._dataTransformation.service import CensusDataTransformer
from the_census._exceptions import NoCensusApiKeyException
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain, SupportedGeoSet
from the_census._geographies.service import GeographyRepository
from the_census._helpers import listAvailableDataSets
from the_census._persistence.interface import ICache
from the_census._persistence.onDisk import OnDiskCache
from the_census._stats.interface import ICensusStatisticsService
from the_census._stats.service import CensusStatisticsService
from the_census._utils.log.configureLogger import DEFAULT_LOGFILE, configureLogger
from the_census._utils.log.factory import ILoggerFactory, LoggerFactory
from the_census._variables.models import GroupCode, VariableCode
from the_census._variables.repository.interface import IVariableRepository
from the_census._variables.repository.models import GroupSet, VariableSet
from the_census._variables.repository.service import VariableRepository
from the_census._variables.search.interface import IVariableSearchService
from the_census._variables.search.service import VariableSearchService

# these are singletons
_serializer = ApiSerializationService()
_transformer = CensusDataTransformer()
_loggerFactory = LoggerFactory()


class Census:
    _client: CensusClient
    _config: Config

    def __init__(
        self,
        year: int,
        datasetType: str = "acs",
        surveyType: str = "acs1",
        cacheDir: str = CACHE_DIR,
        shouldLoadFromExistingCache: bool = False,
        shouldCacheOnDisk: bool = False,
        replaceColumnHeaders: bool = True,
        logFile: str = DEFAULT_LOGFILE,
    ) -> None:
        dotenvPath = dotenv.find_dotenv()

        dotenv.load_dotenv(dotenvPath)

        apiKey = os.getenv("CENSUS_API_KEY")

        if apiKey is None:
            raise NoCensusApiKeyException("Could not find `CENSUS_API_KEY` in .env")

        self._config = Config(
            year,
            datasetType,
            surveyType,
            cacheDir,
            shouldLoadFromExistingCache,
            shouldCacheOnDisk,
            replaceColumnHeaders,
            apiKey,
        )

        container = punq.Container()

        # singletons
        container.register(Config, instance=self._config)
        container.register(ICensusApiSerializationService, instance=_serializer)
        container.register(
            ICensusDataTransformer[pandas.DataFrame], instance=_transformer
        )
        container.register(ILoggerFactory, instance=_loggerFactory)

        # services
        container.register(ICache[pandas.DataFrame], OnDiskCache)
        container.register(ICensusApiFetchService, CensusApiFetchService)
        container.register(IVariableRepository[pandas.DataFrame], VariableRepository)
        container.register(
            IVariableSearchService[pandas.DataFrame], VariableSearchService
        )
        container.register(IGeographyRepository[pandas.DataFrame], GeographyRepository)
        container.register(
            ICensusStatisticsService[pandas.DataFrame], CensusStatisticsService
        )

        # the client
        container.register(CensusClient)

        configureLogger(logFile, datasetName=f"{datasetType}.{surveyType}")

        # for Jupyter
        pandas.set_option("display.max_colwidth", None)  # type: ignore

        self._client = cast(CensusClient, container.resolve(CensusClient))

    # search
    def searchGroups(self, regex: str) -> pandas.DataFrame:
        """
        Searches all group's based on their concept, according
        to `regex`

        Args:
            regex (str)

        Returns:
            pandas.DataFrame: with all of the relevant groups.
        """
        return self._client.searchGroups(regex).copy(deep=True)

    def searchVariables(
        self,
        regex: str,
        *inGroups: GroupCode,
    ) -> pandas.DataFrame:
        """
        - Searches variables based on `regex`.
        - It can search variables based on their name or group concept.
        - Specify `inGroups` with a list of group codes to restrict the search to
        variables within a particular group, or leave it empty to search all variables.
        - This will pull from the API whatever variables aren't in memory.

        Args:
            regex (str)
            inGroups (List[GroupCode], optional): if populated, this will search
            only the variables within the specified groups. Defaults to [].

        Returns:
            pandas.DataFrame: with all of the matched variables
        """
        return self._client.searchVariables(regex, *inGroups).copy(deep=True)

    # repo
    def getGeographyCodes(
        self, forDomain: GeoDomain, *inDomains: GeoDomain
    ) -> pandas.DataFrame:
        """
        Gets geography codes for the specified geography query.
        A GeoDomain is comprised of the domain (e.g., "state"), and an ID
        or wildcard. So passing in a `forDomain` of GeoDomain("congressional district", "*")
        would get all geography codes for all congressional district; providing an `inDomain`
        of, [GeoDomain("state", "06")] would constrain that search to the state with ID
        06 (California).

        Args:
            forDomain (GeoDomain): the primary geography region to query
            inDomains (List[GeoDomain], optional): any parents of the `forDomain`.
            Whether or not these must be populated and/or can have wildcards
            depends on the dataset/survey's supported geographies. (See `supportedGeographies`
            below.). Defaults to [].

        Returns:
            pandas.DataFrame: [description]
        """
        return self._client.getGeographyCodes(forDomain, *inDomains).copy(deep=True)

    def getGroups(self) -> pandas.DataFrame:
        """
        Gets all groups for the dataset/survey

        Returns:
            pandas.DataFrame: with all of the groups
        """
        return self._client.getGroups().copy(deep=True)

    def getVariablesByGroup(self, *groups: GroupCode) -> pandas.DataFrame:
        """
        Gets all variables whose group is in `groups`.

        Args:
            groups (List[GroupCode]): codes of all the groups whose variables you want.

        Returns:
            pandas.DataFrame: with the queried variables.
        """
        return self._client.getVariablesByGroup(*groups).copy(deep=True)

    def getAllVariables(self) -> pandas.DataFrame:
        """
        Gets all variables available to the dataset/survey.
        This may take a while.

        Returns:
            pandas.DataFrame: with all of the variables.
        """
        return self._client.getAllVariables().copy(deep=True)

    def getSupportedGeographies(self) -> pandas.DataFrame:
        """
        Returns a DataFrame with all possible geography query
        patterns for the given dataset/survey.

        Returns:
            pandas.DataFrame
        """
        return self._client.getSupportedGeographies().copy(deep=True)

    def getStats(
        self,
        variablesToQuery: List[VariableCode],
        forDomain: GeoDomain,
        *inDomains: GeoDomain,
    ) -> pandas.DataFrame:
        """
        Gets statistical data based on `variablesToQuery`
        for the specified geographies.

        Args:
            variablesToQuery (List[VariableCode]): the variables to query
            forDomain (GeoDomain)
            inDomains (List[GeoDomain], optional): Defaults to [].

        Returns:
            pandas.DataFrame: with the data
        """
        return self._client.getStats(variablesToQuery, forDomain, *inDomains).copy(
            deep=True
        )

    @staticmethod
    def listAvailabeDatasets() -> pandas.DataFrame:
        """
        The name says it all

        Returns:
            pandas.DataFrame: DataFrame with all available datasets,
            along with their years & descriptions
        """
        return listAvailableDataSets()

    @staticmethod
    def help() -> None:
        print(
            "For more documentation on the census, see https://www2.census.gov/programs-surveys/"
        )
        print(
            "For more documentation on ACS subject defintiions, see https://www2.census.gov/programs-surveys/acs/tech_docs/subject_definitions/2019_ACSSubjectDefinitions.pdf"
        )

    #################################################
    # property variables for Jupyter notebook usage #
    #################################################

    @property
    def variables(self) -> VariableSet:
        return self._client.variables

    @property
    def groups(self) -> GroupSet:
        return self._client.groups

    @property
    def supportedGeographies(self) -> SupportedGeoSet:
        return self._client.supportedGeographies

    def __repr__(self) -> str:
        return f"<Census year={self._config.year} dataset={self._config.datasetType} survey={self._config.surveyType}>"

    def __str__(self) -> str:
        return self.__repr__()



from typing import Collection, List, Set, TypeVar

T = TypeVar("T")


def getUnique(items: Collection[T]) -> List[T]:
    """
    Get all unique elements in a list

    Args:
        items (List[T])

    Returns:
        List[T]: list with distinct items
    """

    uniqueList: List[T] = []
    seenItems: Set[T] = set()

    for item in items:
        if item in seenItems:
            continue
        seenItems.add(item)
        uniqueList.append(item)

    return uniqueList



from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union, cast

import pandas as pd

from the_census._api.models import GeographyItem
from the_census._dataTransformation.interface import ICensusDataTransformer
from the_census._geographies.models import GeoDomain
from the_census._utils.timer import timer
from the_census._variables.models import Group, GroupVariable, VariableCode


class CensusDataTransformer(ICensusDataTransformer[pd.DataFrame]):
    def __init__(self) -> None:
        super().__init__()

    @timer
    def supportedGeographies(
        self, supportedGeos: OrderedDict[str, GeographyItem]
    ) -> pd.DataFrame:
        valuesFlattened: List[Dict[str, str]] = []
        for geoItem in supportedGeos.values():
            for clause in geoItem.clauses:
                valuesFlattened.append(
                    {
                        "name": geoItem.name,
                        "hierarchy": geoItem.hierarchy,
                        "for": clause.forClause,
                        "in": ",".join(clause.inClauses),
                    }
                )

        return pd.DataFrame(valuesFlattened)

    @timer
    def geographyCodes(self, geoCodes: List[List[str]]) -> pd.DataFrame:
        return (
            pd.DataFrame(geoCodes[1:], columns=geoCodes[0])
            .sort_values(by=geoCodes[0][1:])
            .reset_index(drop=True)
        )

    @timer
    def groups(self, groups: Dict[str, Group]) -> pd.DataFrame:
        groupsList = []
        for code, groupObj in groups.items():
            groupDict = {
                "code": code,
                "description": groupObj.description,
                "cleanedName": groupObj.cleanedName,
            }
            groupsList.append(groupDict)

        return pd.DataFrame(groupsList).sort_values(by="code")

    @timer
    def variables(self, variables: List[GroupVariable]) -> pd.DataFrame:
        variableDictList: List[Dict[str, Union[str, int, bool]]] = []

        for variable in variables:
            variableDictList.append(
                {
                    "code": variable.code,
                    "groupCode": variable.groupCode,
                    "groupConcept": variable.groupConcept,
                    "name": variable.name,
                    "predicateType": variable.predicateType,
                    "predicateOnly": variable.predicateOnly,
                    "limit": variable.limit,
                    "cleanedName": variable.cleanedName,
                }
            )

        return (
            pd.DataFrame(variableDictList)
            .sort_values(by=["code"])
            .reset_index(drop=True)
        )

    @timer
    def stats(
        self,
        results: List[List[List[str]]],
        typeConversions: Dict[str, Any],
        geoDomains: List[GeoDomain],
        columnHeaders: Optional[Dict[VariableCode, str]],
    ) -> pd.DataFrame:
        mainDf = pd.DataFrame()
        geoCols = [geoDomain.name for geoDomain in geoDomains]

        for result in results:
            df = pd.DataFrame(result[1:], columns=result[0])

            if mainDf.empty:  # type: ignore
                mainDf = df
            else:
                df = df.drop(columns=["NAME"])  # type: ignore
                mainDf = pd.merge(mainDf, df, on=geoCols, how="inner")  # type: ignore

        allCols = cast(List[str], mainDf.columns.tolist())  # type: ignore

        # reshuffle the columns
        nameCol = ["NAME"]
        variableCols = [col for col in allCols if col != "NAME" and col not in geoCols]

        reorderedColumns = nameCol + geoCols + variableCols

        return (
            mainDf[reorderedColumns]  # type: ignore
            .astype(typeConversions)
            .rename(columns=columnHeaders or {})
            .sort_values(by=geoCols)
            .reset_index(drop=True)
        )



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, Optional, TypeVar

T = TypeVar("T")


class ICache(ABC, Generic[T]):
    """
    Cache for persisting query data,
    so we don't need to make repeat API calls.
    """

    _cachePath: Path

    @abstractmethod
    def put(self, resource: str, data: T) -> bool:
        """
        Adds `data` to the cache

        Args:
            resource (str): string path to identify the resource
            data (T): the data that's being cached

        Returns:
            bool: `True` if the data did not exist in the cache, `False` if it did
        """
        ...

    @abstractmethod
    def get(self, resource: str) -> Optional[T]:
        """
        Gets `resource` from the cache, if it exists, else `None`.

        Args:
            resource (str)

        Returns:
            Optional[T]
        """
        ...

    @property
    def cachePath(self) -> Path:
        return self._cachePath



def cleanVariableName(variableName: str) -> str:
    """
    Removes punctuation/spaces from variable names

    Args:
        variableName (str)

    Returns:
        str: without punctuation
    """

    fullName = ""

    for i, chunk in enumerate(variableName.split("!!")):
        newName = ""
        for subChunk in (
            chunk.replace(",", " ")
            .replace("-", " ")
            .replace(".", " ")
            .replace("(", " ")
            .replace(")", " ")
            .replace("[", " ")
            .replace("]", " ")
            .replace("/", " slash ")
            .split(" ")
        ):
            newName += (
                subChunk.capitalize().replace(" ", "").replace(":", "").replace("'", "")
            )

        if i == 0:
            fullName += newName
        else:
            fullName += f"_{newName}"

    return fullName



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

from the_census._geographies.models import GeoDomain
from the_census._variables.models import VariableCode

_T = TypeVar("_T")


class ICensusStatisticsService(ABC, Generic[_T]):
    """
    Pulls and massages statistical data from the
    Census API
    """

    @abstractmethod
    def getStats(
        self,
        variablesToQuery: List[VariableCode],
        forDomain: GeoDomain,
        *inDomains: GeoDomain,
    ) -> _T:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['the_census',
 'the_census._api',
 'the_census._dataTransformation',
 'the_census._geographies',
 'the_census._persistence',
 'the_census._stats',
 'the_census._utils',
 'the_census._utils.log',
 'the_census._variables',
 'the_census._variables.repository',
 'the_census._variables.search']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.2.0,<2.0.0',
 'punq>=0.4.1,<0.5.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'requests>=2.25.1,<3.0.0',
 'tqdm>=4.55.1,<5.0.0']

setup_kwargs = {
    'name': 'the-census',
    'version': '1.2.10',
    'description': 'Tool to get and query US Census data',
    'long_description': '[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\n![Release](https://github.com/drawjk705/the_census/workflows/Release/badge.svg)\n![CI](https://github.com/drawjk705/the_census/workflows/CI/badge.svg)\n\n# The Census\n\nWant to work with US Census data? Look no further.\n\n## Getting started\n\n### View all datasets\n\nIf you you\'re not sure what Census dataset you\'re interested in, the following code will take care of you:\n\n```python\nfrom the_census import Census\n\nCensus.listAvailableDataSets()\n```\n\nThis will present you with a pandas DataFrame listing all available datasets from the US Census API. (This includes only aggregate datasets, as they other types [of which there are very few] don\'t play nice with the client).\n\n### Selecting a dataset\n\nBefore getting started, you need to [get a Census API key](https://api.census.gov/data/key_signup.html), and set the following the environment variable `CENSUS_API_KEY` to whatever that key is, either with\n\n```bash\nexport CENSUS_API_KEY=<your key>\n```\n\nor in a `.env` file:\n\n```bash\nCENSUS_API_KEY=<your key>\n```\n\nSay you\'re interested in the American Community Survey 1-year estimates for 2019. Look up the dataset and survey name in the table provided by `listAvailableDataSets`, and execute the following code:\n\n```python\n>>> from the_census import Census\n>>> Census(year=2019, datasetType="acs", surveyType="acs1")\n\n<Census year=2019 dataset=acs survey=acs1>\n```\n\nThe `dataset` object will now let you query any census data for the the ACS 1-year estimates of 2019. We\'ll now dive into how to query this dataset with the tool. However, if you aren\'t familiar with dataset "architecture", check out [this](#dataset-architecture) section.\n\n### Arguments to `Census`\n\nThis is the signature of `Census`:\n\n```python\nclass Census\n    def __init__(self,\n                 year: int,\n                 datasetType: str = "acs",\n                 surveyType: str = "acs1",\n                 cacheDir: str = CACHE_DIR,        # cache\n                 shouldLoadFromExistingCache: bool = False,\n                 shouldCacheOnDisk: bool = False,\n                 replaceColumnHeaders: bool = True,\n                 logFile: str = DEFAULT_LOGFILE):  # census.log\n        pass\n```\n\n-   `year`: the year of the dataset\n-   `datasetType`: type of the dataset, specified by [`listAvailableDatasets`](#which-dataset)\n-   `surveyType`: type of the survey, specified by [`listAvailableDatasets`](#which-dataset)\n-   `cacheDir`: if you opt in to on-disk caching (more on this below), the name of the directory in which to store cached data\n-   `shouldLoadFromExistingCache`: if you have cached data from a previous session, this will reload cached data into the `Census` object, instead of hitting the Census API when that data is queried\n-   `shouldCacheOnDisk`: whether or not to cache data on disk, to avoid repeat API calls. The following data will be cached:\n    -   Supported Geographies\n    -   Group codes\n    -   Variable codes\n-   `replaceColumnHeaders`: whether or not to replace column header names for variables with more intelligible names instead of their codes\n-   `logFile`: name of the file in which to store logging information\n\n###### A note on caching\n\nWhile on-disk caching is optional, this tool, by design, performs in-memory caching. So a call to `census.getGroups()` will hit the Census API one time at most. All subsequent calls will retrieve the value cached in-memory.\n\n## Making queries\n\n### Supported geographies\n\nGetting the [supported geographies](#supported-geographies) for a dataset as as simple as this:\n\n```python\ncensus.getSupportedGeographies()\n```\n\nThis will output a DataFrame will all possible supported geographies (e.g., if I can query all school districts across all states).\n\n### Geography codes\n\nIf you decide you want to query a particular geography (e.g., a particular school district within a particular state), you\'ll need the [FIPS](https://en.wikipedia.org/wiki/Federal_Information_Processing_Standard_state_code#FIPS_state_codes) codes for that school district and state.\n\nSo, if you\'re interested in all school districts in Colorado, here\'s what you\'d do:\n\n1. Get FIPS codes for all states:\n\n```python\nfrom the_census import GeoDomain\n\ncensus.getGeographyCodes(GeoDomain("state", "*"))\n```\n\n2. Get FIPS codes for all school districts within Colorado (FIPS code `08`):\n\n```python\ncensus.getGeographyCodes(GeoDomain("school district", "*"),\n                         GeoDomain("state", "08"))\n```\n\nNote that geography code queries must follow supported geography guidelines.\n\n### Groups\n\nWant to figure out what groups are available for your dataset? No problem. This will do the trick for ya:\n\n```python\ncensus.getGroups()\n```\n\n...and you\'ll get a DataFrame with all groups for your census.\n\n#### Searching groups\n\n`census.getGroups()` will return a lot of data that might be difficult to slog through. In that case, run this:\n\n```python\ncensus.searchGroups(regex=r"my regex")\n```\n\nand you\'ll get a filtered DataFrame with matches to your regex.\n\n#### Groups autocomplete\n\nIf you\'re working in a Jupyter notebook and have autocomplete enabled, running `census.groups.`, followed by a tab, will trigger an autocomplete menu for possible groups by their name (as opposed to their code, which doesn\'t have any inherent meaning in and of itself).\n\n```python\ncensus.groups.SexByAge   # code for this group\n```\n\n### Variables\n\nYou can either get a DataFrame of variables based on a set of groups:\n\n```python\ncensus.getVariablesByGroup(census.groups.SexByAge,\n                           census.groups.MedianAgeBySex)\n```\n\nOr, you can get a DataFrame with all variables for a given dataset:\n\n```python\ncensus.getAllVariables()\n```\n\nThis second operation, can, however, take a lot of time.\n\n#### Searching variables\n\nSimilar to groups, you can search variables by regex:\n\n```python\ncensus.searchVariables(r"my regex")\n```\n\nAnd, you can limit that search to variables of a particular group or groups:\n\n```python\ncensus.searchVariables(r"my regex", census.groups.SexByAge)\n```\n\n#### Variables autocomplete\n\nVariables also support autocomplete for their codes, as with groups.\n\n```python\ncensus.variables.EstimateTotal_B01001  # code for this variable\n```\n\n(These names must be suffixed with the group code, since, while variable codes are unique across groups, their names are not unique across groups.)\n\n### Statistics\n\nOnce you have the variables you want to query, along with the geography you\'re interested in, you can now make statistics queries from your dataset:\n\n```python\nfrom the_census import GeoDomain\n\nvariables = census.getVariablesForGroup(census.groups.SexByAge)\n\ncensus.getStats(variables["code"].tolist(),\n                GeoDomain("school district", "*"),\n                GeoDomain("state", "08"))\n```\n\n## Dataset "architecture"\n\nUS Census datasets have 3 primary components:\n\n1.  [Groups](#groups)\n2.  [Variables](#variables)\n3.  [Supported Geographies](#supported-geographies)\n\n### Groups\n\nA group is a "category" of data gathered for a particular census. For example, the `SEX BY AGE` group would provide breakdowns of gender and age demographics in a given region in the United States.\n\nSome of these groups\' names, however, are a not as clear as `SEX BY AGE`. In that case, I recommend heading over to the survey in question\'s [technical documentation](https://www2.census.gov/programs-surveys/) which elaborates on what certain terms mean with respect to particular groups. Unfortunately, the above link might be complicated to navigate, but if you\'re looking for ACS group documentation, [here\'s](https://www2.census.gov/programs-surveys/acs/tech_docs/subject_definitions/2019_ACSSubjectDefinitions.pdf) a handy link.\n\n### Variables\n\nVariables measure a particular data-point. While they have their own codes, you might find variables which share the same name (e.g., `Estimate!!:Total:`). This is because each variable belongs to a [group](#group). So, the `Estimate!!:Total` variable for `SEX BY AGE` group is the total of all queried individuals in that group; but the `Estimate!!:Total` variable for `POVERTY STATUS IN THE PAST 12 MONTHS BY AGE` group is the total of queried individuals for _that_ group. (It\'s important when calculating percentages that you work within the same group. So if I want the percent of men in the US, whose total number I got from `SEX BY AGE` I should use the `Estimate!!:Total:` of that group as my denominator, and not the `Estimate!!:Total:` of the `POVERTY STATUS` group).\n\nVariables on their own, however, do nothing. They mean something only when you query a particular [geography](#supported-geographies) for them.\n\n### Supported Geographies\n\nSupported geographies dictate the kinds of queries you can make for a given census. For example, in the ACS-1, I might be interested in looking at stats across all school districts. The survey\'s supported geographies will tell me if I can actually do that; or, if I need to refine my query to look at school districts in a given state or smaller region.\n',
    'author': 'Joel Krim',
    'author_email': 'drawjk705@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/drawjk705/us-stats/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from the_census._variables.models import GroupCode
from the_census._variables.repository.models import GroupSet, VariableSet

T = TypeVar("T")


class IVariableRepository(ABC, Generic[T]):
    """
    Gets and stores all variable metadata (i.e., codes,
    meanings of variables, etc.)
    """

    # these are collections of all variables/groups
    # that have been pulled from the API so far
    _variables: VariableSet
    _groups: GroupSet

    @abstractmethod
    def getGroups(self) -> T:
        ...

    @abstractmethod
    def getVariablesByGroup(self, *groups: GroupCode) -> T:
        ...

    @abstractmethod
    def getAllVariables(self) -> T:
        ...

    @property
    def variables(self) -> VariableSet:
        return self._variables

    @property
    def groups(self) -> GroupSet:
        return self._groups



import shutil
from logging import Logger
from pathlib import Path

import pandas as pd

from the_census._config import Config
from the_census._persistence.interface import ICache
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer

LOG_PREFIX = "[On-Disk Cache]"


class OnDiskCache(ICache[pd.DataFrame]):
    _config: Config
    _logger: Logger

    def __init__(self, config: Config, loggerFactory: ILoggerFactory) -> None:
        self._config = config
        self._logger = loggerFactory.getLogger(__name__)

        self._cachePath = Path(
            f"{config.cacheDir}/{config.year}/{config.datasetType}/{config.surveyType}"
        )

        if not self._config.shouldCacheOnDisk:
            self._logger.debug("Not creating an on-disk cache")
            return

        self._logger.debug(f"creating cache for {self._cachePath}")

        self.__setUpOnDiskCache()

    @timer
    def __setUpOnDiskCache(self) -> None:
        self._logger.debug("setting up on disk cache")

        if not self._config.shouldLoadFromExistingCache:
            self._logger.debug("purging on disk cache")

            if Path(self._config.cacheDir).exists():
                shutil.rmtree(self._config.cacheDir)

        self._cachePath.mkdir(parents=True, exist_ok=True)

    @timer
    def put(self, resource: str, data: pd.DataFrame) -> bool:
        if not self._config.shouldCacheOnDisk:
            return True

        path = self._cachePath.joinpath(Path(resource))

        if path.exists():
            self._logger.debug(f'resource "{resource}" already exists; terminating')
            return False

        path.parent.mkdir(parents=True, exist_ok=True)

        self._logger.debug(f'persisting "{path}" on disk')

        data.to_csv(str(path.absolute()), index=False)
        return True

    @timer
    def get(self, resource: str) -> pd.DataFrame:
        if (
            not self._config.shouldLoadFromExistingCache
            or not self._config.shouldCacheOnDisk
        ):
            return pd.DataFrame()

        path = self._cachePath.joinpath(Path(resource))

        if not path.exists():
            self._logger.debug(f'cache miss for "{path}"')
            return pd.DataFrame()

        self._logger.debug(f'cache hit for "{path}"')

        return pd.read_csv(path.absolute())  # type: ignore



import logging
from abc import ABC, abstractmethod


class ILoggerFactory(ABC):
    @abstractmethod
    def getLogger(self, name: str) -> logging.Logger:
        ...


class LoggerFactory(ILoggerFactory):
    def __init__(self) -> None:
        super().__init__()

    def getLogger(self, name: str) -> logging.Logger:
        return logging.getLogger(name)
